//
// C++ Implementation: Animation
//
// Description: 
//
//
// Author: Francois Faure <Francois.Faure@imag.fr>, (C) 2008
//
// Copyright: See COPYING file that comes with this distribution
//
//
#include "Animation.h"
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoTransform.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoIndexedFaceSet.h>
#include <Inventor/nodes/SoDrawStyle.h>
#include <Inventor/nodes/SoNodes.h>

#include <iostream>
using std::cerr;
using std::endl;


#include "File_parser.h"
#define PI 3.14159




Animation::Animation( SoGroup* parent )
{

  //********************************************//
  //PARAMETERS                                  //
  //********************************************//

  //number of spatial subdivision
  N_subdiv=4;

  //number of time subdivision
  N_time_subdiv=6;


  //name of file
  const char filename_mesh[1024]={"data/banana_low.g"};
  const char filename_skeleton[1024]={"data/banana_skeleton.dae"};
  const char name_first_node_skeleton[1024]={"1"};

  




  //********************************************//
  //load meshes                                 //
  //********************************************//

  load_meshes(filename_mesh,parent);


  //********************************************//
  //load Skeleton                               //
  //********************************************//

  load_skeleton(filename_skeleton,name_first_node_skeleton);
  create_skeleton_nodes(parent);


  //********************************************//
  //load Animation                              //
  //********************************************//

  //manually load the animation (k_bone,k_axis,A,B,C,omega_1,omega_2,omega_3)
  set_animation(1,0, 1.0,0.3,0.1, 1.5,4.4,11.0);
  set_animation(3,0, -0.9,0.3,0.1, 2.0,7.4,8.0);
  set_animation(3,1, -0.6,0.3,0.1, 1.0,5.4,8.0);
  set_animation(4,2, -0.9,0.3,0.4, 3.0,5.4,10.0);



  //********************************************//
  //Set Distance Threshold                      //
  //********************************************//

  build_distance_threshold();



}

void Animation::update( float t, float )
{

  //local time
  double u=0.0;
  u=(t-int(t));





  //*******************************************************//
  //Find corresponding subdivision index for space and time//
  //*******************************************************//

  //find corresponding multiresolution in space
  //******************************************//
  
  //camera position
  V_3D cam_position(camera->position.getValue()[0],camera->position.getValue()[1],camera->position.getValue()[2]);

  //distance to the mesh
  double distance_to_mesh=cam_position.norm();

  int k_space_threshold=0;
  while(k_space_threshold<N_subdiv-1 && distance_to_mesh<space_threshold[k_space_threshold])
    k_space_threshold++;


  //find corresponding multiresolution in time
  //******************************************//
  int k_time_threshold=0;
  while(k_time_threshold<N_time_subdiv-1 && distance_to_mesh<time_threshold[k_time_threshold])
    k_time_threshold++;

  //set subdivision number
  int current_subdiv=k_space_threshold;//current subdivision level
  int current_time_subdiv=N_time_subdiv-k_time_threshold-1;







  //*******************************************************//
  //enable the drawing of only one subdivided mesh         //
  //*******************************************************//

  for(int k=0;k<N_subdiv;k++)
    style_meshes[k]->style=SoDrawStyleElement::INVISIBLE;    
  style_meshes[current_subdiv]->style=SoDrawStyleElement::FILLED;
  





  //*******************************************************//
  //SKINNING                                               //
  //*******************************************************//

  //get skeleton to deform
  deformed_skeleton=skinned_vector[current_subdiv].get_deformed_skeleton();
  
  //deform it
  skeleton_vector[current_time_subdiv].deform_skeleton(u,deformed_skeleton);

  //set this deformed skeleton into the skinned class
  skinned_vector[current_subdiv].set_deformed_skeleton(deformed_skeleton);

  //perform the skinng
  skinned_vector[current_subdiv].skinning();

  //copy the result into the graph
  V_3D x;
  for(int k=0;k<skinned_vector[current_subdiv].get_mesh()->get_vertex_number();k++)
    {
      x=skinned_vector[current_subdiv].get_deformed_mesh()->get_vertex_v3d(k);
      coordinates_meshes[current_subdiv]->point.set1Value(k,x[0],x[1],x[2]);
    }






  //*******************************************************//
  //Update Skeleton Drawing                                //
  //*******************************************************//


  Matrix M;
  SbMatrix SoM;

  //just to see the skeleton
  SbMatrix to_translate(1,0,0,0, 0,1,0,0, 0,0,1,0, 0.2,0,0,1); 

  for(int k_joint=0;k_joint<skeleton_vector[current_subdiv].get_N_joint();k_joint++)
    {
      M=deformed_skeleton->get_joint(k_joint)->get_world_matrix();

      //copy matrix
      for(int k1=0;k1<4;k1++)
	for(int k2=0;k2<4;k2++)
	  SoM[k1][k2]=M.v(k1,k2);

      //X axis
      skeleton_transform[3*k_joint+0]->rotation.setValue(SbVec3f(0.0,0.0,1.0),PI/2.0);
      skeleton_transform[3*k_joint+0]->translation.setValue(0.05/2.0,0.0,0.0);
      skeleton_transform[3*k_joint+0]->multRight(SoM);
      skeleton_transform[3*k_joint+0]->multRight(to_translate);

      //Y axis
      skeleton_transform[3*k_joint+1]->rotation.setValue(SbVec3f(0.0,1.0,0.0),0.0);
      skeleton_transform[3*k_joint+1]->translation.setValue(0.0,0.05/2.0,0.0);
      skeleton_transform[3*k_joint+1]->multRight(SoM);
      skeleton_transform[3*k_joint+1]->multRight(to_translate);

      //Z axis
      skeleton_transform[3*k_joint+2]->rotation.setValue(SbVec3f(1.0,0.0,0.0),PI/2.0);
      skeleton_transform[3*k_joint+2]->translation.setValue(0.0,0.0,0.05/2.0);
      skeleton_transform[3*k_joint+2]->multRight(SoM);
      skeleton_transform[3*k_joint+2]->multRight(to_translate);

    }



}



Animation::~Animation()
{
}



int Animation::set_camera(SoCamera *_camera){camera=_camera;return 0;}


//set-up Multiresolution in time for every skeleton_vector
int Animation::build_time_multires(const std::vector <double> angle,std::vector <double_vector> *average,std::vector<double_vector> *residual)
{
  average->resize(N_time_subdiv);
  residual->resize(N_time_subdiv);
  
  int k_subdiv=0;
  int k_time=0;
  int current_size=angle.size();
  (*average)[0].set(angle);


  //MULTIRESOLUTION                                  //
  //*************************************************//
  for(k_subdiv=1;k_subdiv<N_time_subdiv;k_subdiv++)
    {
      current_size/=2;
      for(k_time=0;k_time<current_size;k_time++)
	{
	  (*average)[k_subdiv].add(
				   0.5*
				   (
				    (*average)[k_subdiv-1][2*k_time+0]
				    +
				    (*average)[k_subdiv-1][2*k_time+1]
				    )
				   );
	  (*residual)[k_subdiv].add(
				    (*average)[k_subdiv-1][2*k_time+1]
				    -
				    (*average)[k_subdiv-1][2*k_time+0]
				    );
	}

    }
  



  //symmetrize for animation
  //*************************************************//
  for(k_subdiv=0;k_subdiv<N_time_subdiv;k_subdiv++)
    {
      current_size=(*average)[k_subdiv].size();
      for(k_time=0;k_time<current_size;k_time++)
	(*average)[k_subdiv].add((*average)[k_subdiv][current_size-k_time-1]);
    }
      
  
  


  return 0;
}


//load meshes, subdivided them and create the graph
int Animation::load_meshes(const char* filename,SoGroup *parent)
{

  File_parser parser;
  int k_subdiv=0;
  int k=0;
  V_3D x;

  separator_meshes   .resize(N_subdiv);
  coordinates_meshes .resize(N_subdiv);
  index_face_meshes  .resize(N_subdiv);
  style_meshes       .resize(N_subdiv);
  skinned_vector     .resize(N_subdiv);


  Mesh temp_mesh;
  std::vector <int> temp_connectivity;
  for(k_subdiv=0;k_subdiv<N_subdiv;k_subdiv++)
    {


      //Allocate nodes
      //**************************//
      separator_meshes  [k_subdiv]=new SoSeparator;
      coordinates_meshes[k_subdiv]=new SoCoordinate3;
      index_face_meshes [k_subdiv]=new SoIndexedFaceSet;
      style_meshes      [k_subdiv]=new SoDrawStyle;





      //load meshes and create spatial subdivision
      //**************************//      
      if(k_subdiv==0)//special case (seg fault if subdivide_mesh is called with 0 subdivision)
	parser.load_g_mesh(filename,&temp_mesh);
      else
	subdivide_mesh(2,k_subdiv,filename,&temp_mesh);




      //load mesh into the skinning class
      //**************************//
      *(skinned_vector[k_subdiv].get_mesh()) = temp_mesh;
      skinned_vector[k_subdiv].get_mesh()->build_smooth_normal();




      //load coordinates into node
      //**************************//
      temp_connectivity=temp_mesh.get_connectivity_vrml();
      for(k=0;k<temp_mesh.get_vertex_number();k++)
	{
	  x=temp_mesh.get_vertex_v3d(k);
	  coordinates_meshes[k_subdiv]->point.set1Value(k,x[0],x[1],x[2]);
	}



      //load connectivity into node
      //**************************//
      for(k=0;k<int(temp_connectivity.size());k++)
	index_face_meshes[k_subdiv]->coordIndex.set1Value(k,temp_connectivity[k]);
      



      //associate nodes
      //**************************//
      separator_meshes[k_subdiv]->addChild(style_meshes[k_subdiv]);
      separator_meshes[k_subdiv]->addChild(coordinates_meshes[k_subdiv]);
      separator_meshes[k_subdiv]->addChild(index_face_meshes[k_subdiv]);
 
      
      parent->addChild(separator_meshes[k_subdiv]);
    }

  
  return 0;
}


int Animation::load_skeleton(const char* filename,const char* name_first_node)
{

  File_parser parser;

  //load skeleton
  Skeleton temp_skeleton;
  parser.load_collada_skeleton(filename,name_first_node,&temp_skeleton);


  //set matrices
  temp_skeleton.fix_init_bind_pose();
  temp_skeleton.read_skeleton();
  temp_skeleton.precalculate_matrix();



  //skinning weight
  int k_subdiv=0;
  for(k_subdiv=0;k_subdiv<N_subdiv;k_subdiv++)
    {
      skeleton=new Skeleton;
      deformed_skeleton=new Skeleton;

      *skeleton=temp_skeleton;
      *deformed_skeleton=temp_skeleton;

      skinned_vector[k_subdiv].set_skeleton(skeleton);
      skinned_vector[k_subdiv].set_deformed_skeleton(deformed_skeleton);
      skinned_vector[k_subdiv].automatic_weight_calculation();
    }





  //set-up skeleton size
  //**************************//

  skeleton_vector.resize(N_time_subdiv);

  int current_index=0;
  int k_time_subdiv=0;
  int max_size=2*int(powf(2.0,double(N_time_subdiv+1)));//2* => symetric
  int current_size=0;

  //this size is divided by two at each multiresolution level
  current_size=max_size;

  for(k_time_subdiv=0;k_time_subdiv<N_time_subdiv;k_time_subdiv++)
    {
      //init skeleton
      skeleton_vector[k_time_subdiv]=temp_skeleton;

      //set animation size, get ready for multiresolution
      for(int k=0;k<skeleton_vector[current_index].get_N_joint();k++)
	{
	  skeleton_vector[k_time_subdiv].set_size_time(k,current_size);
	  for(int k_motion=0;k_motion<current_size;k_motion++)
	    skeleton_vector[k_time_subdiv].set_time(k,k_motion,double(k_motion)/double(current_size-1.0));
	}
      current_size/=2;//multiresolution size
    }



  return 0;
}




//load animation of every skeleton and pre-compute subdivision
int Animation::set_animation(int k_bone,int k_dim,double A,double B,double C,double omega_1,double omega_2,double omega_3)
{

  int k=0,k_time_subdiv=0;
  int N_motion=int(powf(2.0,double(N_time_subdiv+1)));





  //vector of angle at the best resolution
  std::vector <double> motion_angle(N_motion);

  //multiresolution functions
  std::vector <double_vector> multires_angle(N_motion* N_time_subdiv);
  std::vector <double_vector> multires_angle_residual(N_motion* N_time_subdiv);



  //***************************************//
  //build angle function                   //
  //                                       //
  //CAN BE ANY FUNCTIONS                   //
  //***************************************//
  double u=0.0;
  for(k=0,u=0.0;k<N_motion;k++,u=double(k)/double(N_motion-1))
    motion_angle[k] = 20*(A*sin(omega_1*u*PI)+B*cos(omega_2*u*PI)+C*sin(omega_3*u*PI));





  //build multiresolution of angle vector
  //*******************************//
  build_time_multires(motion_angle,&multires_angle,&multires_angle_residual);




  //fill skeletons with animation
  //*******************************//
  for(k_time_subdiv=0;k_time_subdiv<N_time_subdiv;k_time_subdiv++)
    {
      int size_motion=multires_angle[k_time_subdiv].size();

      //fill the multiresolution angle
      for(int k_motion=0;k_motion<size_motion;k_motion++)
	skeleton_vector[k_time_subdiv].set_angle(k_bone,k_motion,k_dim,multires_angle[k_time_subdiv][k_motion]);

    }

  return 0;
}

int Animation::create_skeleton_nodes(SoGroup* parent)
{
  //Add Skeleton in scene Graph

  int k_joint=0,k_dim=0;
  int N_joint=skeleton_vector[0].get_N_joint();
  separator_skeleton.resize(3*N_joint);
  cylinder_vector.resize(3*N_joint);
  skeleton_transform.resize(3*N_joint);
  for(k_joint=0;k_joint<N_joint;k_joint++)
    {
      for(k_dim=0;k_dim<3;k_dim++)
	{
	  separator_skeleton[3*k_joint+k_dim]=new SoSeparator;
	  parent->addChild(separator_skeleton[3*k_joint+k_dim]);

	  skeleton_transform[3*k_joint+k_dim]=new SoTransform;
	  separator_skeleton[3*k_joint+k_dim]->addChild(skeleton_transform[3*k_joint+k_dim]);

	  cylinder_vector[3*k_joint+k_dim]=new SoCylinder;
	  cylinder_vector[3*k_joint+k_dim]->radius.setValue(0.005);
	  cylinder_vector[3*k_joint+k_dim]->height.setValue(0.05);
	  separator_skeleton[3*k_joint+k_dim]->addChild(cylinder_vector[3*k_joint+k_dim]);
	}
    }


  return 0;
}



//distance threshold for subdivision
int Animation::build_distance_threshold()
{
  
  //space threshold
  space_threshold.resize(N_subdiv-1);
  space_threshold[N_subdiv-2]=0.7;
  for(int k=N_subdiv-3;k>=0;k--)
    space_threshold[k]=space_threshold[k+1]*2.0;

  //set distance threshold for time parameter
  time_threshold.resize(N_time_subdiv);
  for(int k=0;k<N_time_subdiv-1;k++)
    {
      //linear cut
      double u=double(k)/double(N_time_subdiv-1.0);
      time_threshold[k]=4*space_threshold[0]*(1.0-u)+space_threshold[N_subdiv-2]*u;
    }

  
  return 0;

}
