//
// C++ Interface: utils
//
// Description: Various tools for openinventor scenes
//
//
// Author: Francois Faure <Francois.Faure@imag.fr>, (C) 2005
//
// Copyright: See COPYING file that comes with this distribution
//
//
#pragma once

#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/SbLinear.h>

#include <iostream>
#include <vector>

namespace oi
{
namespace utils
{

/// write the scene to a given file
void writeFile( SoGroup* root, const char* filename, bool ascii=true );
/** Read file with the given name. Return 0 and output an error message if not succeeded */
SoSeparator *readFile(const char *filename, const char* nodename="");
/** Load a file and append it as child of the given node. Output an error message if not succeeded.*/
void appendFile( SoGroup* groupNode, const char* fileName );

///// write an SbVec3f to an output flow
//std::ostream& operator << (std::ostream& out, const SbVec3f& v );
/// read an SbVec3f from an input flow
std::istream& operator >> (std::istream& in, SbVec3f& v );

/// dot product
float operator * ( const SbVec3f& a, const SbVec3f& b );

/// write an std::vector to an output flow
template<class T>
inline std::ostream& operator << (std::ostream& out, const std::vector<T>& v );
template<class T>
inline std::ostream& operator << (std::ostream& out, const std::vector<T>& v )
{
    for( typename std::vector<T>::const_iterator i=v.begin(), iend=v.end(); i!=iend; ++i)
        out << *i <<", ";
    return out;
}

inline std::ostream& operator << (std::ostream& out, SbString sbstring )
{
    out<<sbstring.getString();
    return out;
}

/// Create three lines to represent the axes of the reference frame.
SoSeparator* viewFrame();

/// Cherche un noeud par son nom sous le noeud pass� en param�tre. Renvoie le chemin complet de la racine vers le noeud trouv�. Renvoie null si pas trouv�. Si il y en a plusieurs, renvoie le premier trouv�.
SoPath* findPathByName( SoNode* tree, char* name);

/// Write in ascii the given path to the given file
void writePath( SoPath* p, FILE* fp=stderr );

/// Calls SoNode::getByName and casts it to the template parameter type. Return 0 if the node is not found or if it is of the wrong type.
template<class T>
inline T* findNode( char* name )
{
    //cerr<<"findNode found node named "<< name;
    SoNode* node = SoNode::getByName( SbName(name) );
    //cerr << " of type "<< node->getTypeId().getName().getString() << endl;
    if( node && node->isOfType(T::getClassTypeId()) )
        return (T*) node;
    else
        return 0;
}

/// Set all the values of an SoMFField
template<typename MField, typename SbValue>
inline void setAllValues( MField& mf, const SbValue& v )
{
    SbValue* array = mf.startEditing();
    unsigned num = mf.getNum();
    for( unsigned i=0; i<num; ++i )
        array[i] = v;
    mf.touch();
}

/// Return the name of the node, or its type if the name is empty
SbString nameOrType ( SoBase* b );

/** Looks for the first node with the given name among the children of the given group. The search is not recursive. Return 0 if not found */
SoNode* findChild( SoGroup*, SbString name);

/** Looks for the first child of the given group with the given type. Not recursive. Return 0 if not found */
SoNode* findChild( SoGroup*, SoType type );

/** Looks for the first child of the given group with the given type and the given name. Not recursive. Return 0 if not found */
SoNode* findChild( SoGroup*, SoType type, SbString name );

/** remove the children of a group */
void clearGroup( SoGroup * );

}//utils
}//oi

