//
// C++ Implementation: utils
//
// Description:
//
//
// Author: Francois Faure <Francois.Faure@imag.fr>, (C) 2005
//
// Copyright: See COPYING file that comes with this distribution
//
//

//#include "stdafx.h"
#include "oi.h"
#include <stdio.h>
#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/SoInput.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoMaterialBinding.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoLineSet.h>
#include <iostream>
using std::cout;
using std::cerr;
using std::endl;

namespace oi
{
namespace utils
{

/** Read the file with given name and returns a pointer to the subgraph created.
Return Null if the graph could not be built.
*/
SoSeparator *
readFile(const char *filename, const char* nodename )
{
    //cerr<<"readFile "<<filename<<endl;
    // Open the input file
    SoInput mySceneInput;
    if (!mySceneInput.openFile(filename))
    {
        cerr<<"Cannot open file "<<filename<<endl;
        return NULL;
    }

    // Read the whole file into the database
    SoSeparator *myGraph = SoDB::readAll(&mySceneInput);
    if (myGraph == NULL)
    {
        cerr<<"Could not build graph from file "<<filename<<endl;
        cerr<<"Check if the format confroms to OpenInventor "<<endl;
        return NULL;
    }

    mySceneInput.closeFile();
    
    if( strlen(nodename)>1 )
        myGraph->setName(nodename);
    return myGraph;
}

void appendFile( SoGroup* groupNode, const char* fileName )
{
    if( SoSeparator* tree = readFile(fileName) )
        groupNode->addChild( tree );
}

/** Write the graph to a file.
\param root root of the graph to write
\param filename name of the written file
\param ascii if true write in ascii else write in binary format
*/
void writeFile( SoGroup* root, const char* filename, bool ascii )
{
    SoOutput out;
    if( !out.openFile(filename) )
    {
        cerr<<"Cannot open file "<<filename<<" for writing"<<endl;
    }
    SoWriteAction myWriteAction(&out);
    if( ascii )
        myWriteAction.getOutput()->setBinary( false );
    else
        myWriteAction.getOutput()->setBinary( true );
    myWriteAction.apply(root);
}



///// write a SbVec3f to an output flow
//std::ostream& operator << (std::ostream& out, const SbVec3f& v )
//{
//    out << v[0] <<" "<< v[1] <<" "<< v[2];
//    return out;
//}

/// read a SbVec3f from an input flow
std::istream& operator >> (std::istream& in, SbVec3f& v )
{
    in >> v[0] >> v[1] >> v[2];
    return in;
}

float operator * ( const SbVec3f& a, const SbVec3f& b )
{
    return a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
}

/// Create three lines to represent the axes of the reference frame.
SoSeparator* viewFrame()
{
    SoSeparator* frame = new SoSeparator;
    SoMaterial* material = new SoMaterial;
    frame->addChild( material );
    float colors[6][3] = {{1,0,0}, {1,0,0}, {0,1,0}, {0,1,0}, {0,0,1}, {0,0,1}};
    material->diffuseColor.setValues(0,6,colors);
    SoMaterialBinding* materialBinding = new SoMaterialBinding;
    frame->addChild( materialBinding );
    materialBinding->value.setValue(SoMaterialBindingElement::PER_VERTEX);
    SoCoordinate3* coords = new SoCoordinate3;
    frame->addChild(coords);
    float x[6][3]={{0,0,0}, {1,0,0}, {0,0,0}, {0,1,0}, {0,0,0}, {0,0,1}};
    coords->point.setValues(0,6,x);
    SoLineSet* lines = new SoLineSet;
    frame->addChild(lines);
    int span[3] = {2,2,2};
    lines->numVertices.setValues(0,3,span);
    return frame;

}

/// Cherche un noeud par son nom sous le noeud pass� en param�tre. Renvoie le chemin complet de la racine vers le noeud trouv�. Renvoie null si pas trouv�. Si il y en a plusieurs, renvoie le premier trouv�. 
SoPath* findPathByName( SoNode* tree, char* name)
{
    SoSearchAction search;
    search.setName(name);
    search.setInterest(SoSearchAction::FIRST);
    search.apply(tree);
    if( !search.getPath() ) return 0;
/*    if( search.getPath()->getLength()==0 ) 
        return 0;*/
    return search.getPath()->copy();
}

void writePath( SoPath* p, FILE* fp )
{
    SoWriteAction write;
    write.getOutput()->setFilePointer(fp);
    write.apply(p);
}

SbString nameOrType ( SoBase* b )
{
    SbString name = b->getName().getString();
    if( name.getLength()==0 )
        return b->getTypeId().getName().getString();
    else
        return name;
}

SoNode* findChild( SoGroup* g, SbString name)
{
    for( int i=0; i<g->getNumChildren(); ++i )
    {
        if( g->getChild(i)->getName().getString()==name )
            return g->getChild(i);
    }
    return 0;
}

SoNode* findChild( SoGroup* g, SoType type )
{
    for( int i=0; i<g->getNumChildren(); ++i )
    {
        if( g->getChild(i)->isOfType(type) )
            return g->getChild(i);
    }
    return 0;
}

SoNode* findChild( SoGroup* g, SoType type, SbString name )
{
    SoNode* node = findChild(g,name);
    if( node && node->isOfType(type) )
        return node;
    return 0;
}

void clearGroup( SoGroup* g )
{
    while( g->getNumChildren() > 0 )
        g->removeChild( g->getChild(0) );
}



}
}
